from lark import Lark, Tree, Token
from mistql.expression import (
    RefExpression,
    FnExpression,
    ValueExpression,
    ArrayExpression,
    ObjectExpression,
    PipeExpression,
)
from typing import Union, List, Any
import json

from mistql.expression import BaseExpression
from mistql.exceptions import OpenAnIssueIfYouGetThisError


mistql_parser = Lark.open("grammar.lark", rel_to=__file__, parser="earley")

function_mappings = {
    "dot": ".",
    "neg": "-/unary",
    "not": "!/unary",
    "gt": ">",
    "lt": "<",
    "gte": ">=",
    "lte": "<=",
    "eq": "==",
    "neq": "!=",
    "match": "=~",
    "plus": "+",
    "minus": "-",
    "mul": "*",
    "div": "/",
    "mod": "%",
    "and": "&&",
    "or": "||",
}


def process_lark_tree(lark_node: Tree) -> BaseExpression:
    if lark_node.data == "array":
        return ArrayExpression([from_lark(child) for child in lark_node.children])
    elif lark_node.data == "object":
        obj = {}
        for child in lark_node.children:
            if isinstance(child, str):
                raise OpenAnIssueIfYouGetThisError(
                    "Got string for child when we didn't expect it."
                )
            key = from_lark(child.children[0])
            if isinstance(key, ValueExpression):
                key = key.value.value
            elif isinstance(key, RefExpression):
                key = key.name
            else:
                raise OpenAnIssueIfYouGetThisError(f"Unknown key type {type(key)}")
            value = from_lark(child.children[1])
            obj[key] = value
        return ObjectExpression(obj)
    elif lark_node.data == "pipe":
        return PipeExpression([from_lark(child) for child in lark_node.children])
    elif lark_node.data == "fncall":
        return FnExpression(
            from_lark(lark_node.children[0]),
            [from_lark(child) for child in lark_node.children[1:]],
        )
    elif lark_node.data in function_mappings:
        return FnExpression(
            RefExpression(function_mappings[lark_node.data], absolute=True),
            [from_lark(child) for child in lark_node.children[:]],
        )
    elif lark_node.data == "index":
        # This is gross becase i can't figure out how to get the tree to look
        # a little more sensible.
        base, indexing = lark_node.children
        if isinstance(indexing, str):
            raise OpenAnIssueIfYouGetThisError(
                "Got string for child when we didn't expect it."
            )
        innards = indexing.children[0]
        if isinstance(innards, str):
            raise OpenAnIssueIfYouGetThisError(
                "Got string for child when we didn't expect it."
            )
        fnexp_args: List[BaseExpression] = []
        prev_was_token = True
        for child in innards.children:
            if isinstance(child, Token) and child.value == ":":
                if prev_was_token:
                    fnexp_args.append(ValueExpression.of(None))
                prev_was_token = True
                continue
            else:
                fnexp_args.append(from_lark(child))
                prev_was_token = False
        if prev_was_token:
            fnexp_args.append(ValueExpression.of(None))
        fnexp_args.append(from_lark(base))
        return FnExpression(RefExpression("index", absolute=True), fnexp_args)
    else:
        raise OpenAnIssueIfYouGetThisError(
            f"Unknown lark expression type: {lark_node.data}"
        )


def process_lark_token(lark_node: Token) -> BaseExpression:
    if lark_node.type == "NUMBER":
        return ValueExpression.of(float(lark_node.value))
    elif lark_node.type == "ESCAPED_STRING":
        value = json.loads(lark_node.value)
        return ValueExpression.of(value)
    elif lark_node.type == "TRUE":
        return ValueExpression.of(True)
    elif lark_node.type == "FALSE":
        return ValueExpression.of(False)
    elif lark_node.type == "NULL":
        return ValueExpression.of(None)
    elif lark_node.type == "CNAME":
        return RefExpression(lark_node.value)
    elif lark_node.type == "AT":
        return RefExpression("@")
    elif lark_node.type == "DOLLAR":
        return RefExpression("$")
    else:
        raise OpenAnIssueIfYouGetThisError(f"Unknown token type {lark_node.type}")


def from_lark(lark_node: Union[Any, str, Tree, Token]):
    if isinstance(lark_node, Token):
        return process_lark_token(lark_node)
    elif isinstance(lark_node, Tree):
        return process_lark_tree(lark_node)
    else:
        raise OpenAnIssueIfYouGetThisError(f"Unknown lark node type: {type(lark_node)}")


def parse(raw):
    # TODO: Translate errors from this function to something that inherits
    # from MistQLException
    parsed = mistql_parser.parse(raw)
    return from_lark(parsed)



from typing import Any

from .execute import execute_outer
from .gardenwall import input_garden_wall, output_garden_wall
from .parse import parse


def query(query: str, raw_data: Any) -> Any:
    """
    Executes a query on a given data.

    :param query: The query to execute.
    :param data: The data to query.
    :return: The result of the query.
    """
    ast = parse(query)
    data = input_garden_wall(raw_data)
    result = execute_outer(ast, data)
    return_value = output_garden_wall(result)
    return return_value



from typing import List
from mistql.runtime_value import RuntimeValue, RuntimeValueType
from mistql.expression import (
    Expression,
    RefExpression,
    FnExpression,
    ValueExpression,
    ArrayExpression,
    ObjectExpression,
    PipeExpression,
)
from mistql.stack import Stack
from mistql.builtins import FunctionDefinitionType, builtins
from mistql.stack import add_runtime_value_to_stack, build_initial_stack, find_in_stack
from mistql.expression import BaseExpression
from mistql.exceptions import MistQLTypeError, OpenAnIssueIfYouGetThisError

from typeguard import typechecked


@typechecked
def execute_fncall(head: BaseExpression, arguments: List[BaseExpression], stack: Stack):
    fn = execute(head, stack)
    if fn.type != RuntimeValueType.Function:
        raise MistQLTypeError(f"Tried to call a non-function: {fn}")
    # Not enforced, but definitely should be.
    function_definition: FunctionDefinitionType = fn.value
    return function_definition(arguments, stack, execute)


@typechecked
def execute_pipe(stages: List[BaseExpression], stack: Stack) -> RuntimeValue:
    first: BaseExpression = stages[0]
    remaining: List[BaseExpression] = stages[1:]
    data = execute(first, stack)

    for stage_ast in remaining:
        new_stack = add_runtime_value_to_stack(data, stack)
        fn: BaseExpression
        args: List[BaseExpression]
        if isinstance(stage_ast, FnExpression):
            fn = stage_ast.fn
            args = stage_ast.args.copy()
        else:
            fn = stage_ast
            args = []
        args.append(ValueExpression(data))
        stage = FnExpression(fn, args)
        data = execute(stage, new_stack)

    return data


@typechecked
def execute(ast: BaseExpression, stack: Stack) -> RuntimeValue:
    if not isinstance(ast, BaseExpression):
        raise OpenAnIssueIfYouGetThisError(
            f"Expected to evaluate an expression, got {ast}"
        )
    if isinstance(ast, ValueExpression):
        return ast.value
    elif isinstance(ast, RefExpression):
        return find_in_stack(stack, ast.name, ast.absolute)
    elif isinstance(ast, FnExpression):
        return execute_fncall(ast.fn, ast.args, stack)
    elif isinstance(ast, ArrayExpression):
        return RuntimeValue.of([execute(item, stack) for item in ast.items])
    elif isinstance(ast, ObjectExpression):
        return RuntimeValue.of(
            {key: execute(value, stack) for key, value in ast.entries.items()}
        )
    elif isinstance(ast, PipeExpression):
        return execute_pipe(ast.stages, stack)
    raise NotImplementedError("execute() not implemented for " + str(ast.type))


def execute_outer(ast: Expression, data: RuntimeValue) -> RuntimeValue:
    return execute(ast, build_initial_stack(data, builtins))



from typing import Callable, List, Dict
from mistql.runtime_value import RuntimeValue, RuntimeValueType
from mistql.exceptions import MistQLReferenceError
from typeguard import typechecked

StackFrame = Dict[str, RuntimeValue]
Stack = List[StackFrame]


def add_runtime_value_to_stack(value: RuntimeValue, stack: Stack):
    new_stackframe = {"@": value}
    for key in value.keys():
        new_stackframe[key] = value.access(key)
    new_stack = stack.copy()
    new_stack.append(new_stackframe)
    return new_stack


def build_initial_stack(data: RuntimeValue, builtins: Dict[str, Callable]) -> Stack:
    dollar_var = {
        "@": data,
    }
    top_stack_entry = {
        "@": data,
    }
    if data.type == RuntimeValueType.Object:
        for key, value in data.value.items():
            top_stack_entry[key] = value
    for builtin in builtins:
        fn = RuntimeValue.create_function(builtins[builtin])
        dollar_var[builtin] = fn
    top_stack_entry["$"] = RuntimeValue.of(dollar_var)
    builtin_stackframe = {
        name: RuntimeValue.create_function(builtins[name]) for name in builtins
    }
    return [builtin_stackframe, top_stack_entry]


@typechecked
def find_in_stack(stack: Stack, name: str, absolute: bool) -> RuntimeValue:
    if absolute:
        stack = stack[:1]
    for frame in reversed(stack):
        if name in frame:
            return frame[name]
    raise MistQLReferenceError(f"Could not find {name} in stack")
